/**************************************************************

   Program Title:  P1 AVR Boot/Monitor
   
   Program Name:  P1_AVR_SPI_Flash.c	Version: 0.88
   
   Author:  SS	  					  Date:  August 10, 2005
   
   Compiled with ImageCraft ICCAVR Standard V7.02
   
   The AVR SPI lines are connected to two serial flash devices U2 
   and U5.  The Blackfin boots from U2 when set to boot mode 3 as 
   an SPI Master.  If the AVR needs to communicate with the Serial 
   Flash devices the Blackfin must relinquish control of the SPI 
   lines.  This version of the code does not support cooperative sharing
   of the SPI lines so the Blackfin is reset and forced to boot under 
   mode 0 freeing the SPI lines.
   
   Serial flash devices compatible with the P1 are the AT45DB041
   (512kx8), AT45DB081 (1Mx8), AT45DB161 (2Mx8), AT45DB321 (4Mx8)
   or AT45DB642 (8Mx8).
   
   The Serial Flash chips are page oriented with 2 SRAM buffers of 264,
   528 or 1024 bytes each.  The AVR programs the serial flash devices by 
   loading files via the serial line (in IntelHex32 format) into the RAM 
   buffers before transfer to the appropriate page.  Page erase/programming 
   requires about 30msec per page. The routines in this file perform these 
   functions.  The host terminal program must support XON/XOFF flow control.
   
   Copyright 2004-2007, SOC Robotics, Inc.  All rights reserved.
   
****************************************************************/
/*
 * Copyright (C) 2007 by SOC Robotics, Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 * 4. Redistribution is only allowed in combination with SOC Machines, 
 *    Inc. hardware products.
 *
 * THIS SOFTWARE IS PROVIDED BY SOC ROBOTICS, INC. AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL OC MACHINES, 
 * INC. OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.soc-machines.com
 */

/*
 * $Log: P1_Monitor.c,v $
 * Revision 0.88  2007/07/23 16:40:20  SS
 *    2-wire interface, 8MHz clock, Serial Flash Writer, XON/XOFF
 * Revision 0.85  2006/01/08 16:40:20  SS
 *    2-wire interface, 8MHz clock, Serial Flash Writer, XON/XOFF
 * Revision 0.52  2005/06/04 18:36:42  SS
 * Preview release
 *
 */
#include <iom644v.h>
#include <macros.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include "P1_AVR_Uart.h"
#include "P1_AVR_Defines.h"
#include "P1_AVR_Extern.h"

// Set Serial Flash chip select state
void FlashChipSelect(unsigned char chipno,unsigned char mode)
{
 	 if(chipno==FLASHCHIP_1) {
	 	if(mode==LOW) PORTC &= 0xEF; 
		else PORTC |= 0x10;  
		}
	 else if(chipno==FLASHCHIP_2) {
	 	if(mode==LOW) PORTD &= 0xFB; 
		else PORTD |= 0x04;  
	    }
}

//  SPI Flash device address construction function
//  Read address consists of 4 zero bits+11 page bits+9 byte start bits
void ConvertPower2FlashAddress(long address)
{
    FlashAddress[0] = (address>>16)&0x00FF;
    FlashAddress[1] = (address>>8)&0x00FF;
    FlashAddress[2] = address&0x00FF;
}

//  SPI Flash device address construction function
//  Read address consists of 4 zero bits+11 page bits+9 byte start bits
void ConvertFlashAddress(long address)
{
    unsigned char inchar;
 	unsigned int pages,bytestart,superpages;
	
	//  Use simple addressing if 321D or 642D
	if((flashpagesize0==512)||(flashpagesize0==1024))  
	                    ConvertPower2FlashAddress(address);
	else {
//	    putchar('=');
//	    PrintLongHex(address,6);
//	    putchar(' ');	
	    pages = address/flashpagesize0;
	    superpages = pages/256;
	    bytestart = address - (long)pages*flashpagesize0;
	    inchar = ((superpages&0x00ff)<<1) + ((pages&0x00ff)>>7);
        FlashAddress[0] = inchar;
//	    PrintHex(inchar);
	    inchar = ((pages&0x00ff)<<1) + (bytestart>>8);
	    FlashAddress[1] = inchar;
//	    PrintHex(inchar);
	    inchar = bytestart & 0x00FF;
	    FlashAddress[2] = inchar;
//	    PrintHex(inchar);
//	    putchar('=');	
		}
}

//  Convert Serial Flash to Power of 2 device
//  For AT45DB642D  - send 3D 2A 80 A6
//  For AT45DB321D  - send 3D 2A 80 A6
//  Function assumes Serial Flash is 642 or 321
void ProgramPower2(unsigned char chipno)
{
	 //  Set PC4 or PD2 low - Flash CS
	 FlashChipSelect(chipno,LOW);

	 // Send Power of 2 command sequence
 	 SPI_MasterTransmit(0x3D);
 	 SPI_MasterTransmit(0x2A);
 	 SPI_MasterTransmit(0x80);
 	 SPI_MasterTransmit(0xA6);

	 // Set PC4 or PD2 high - deselect Flash chip
	 FlashChipSelect(chipno,HIGH);
	 
	 Delaymsec(20);
}

//  Get serial flash chip status
unsigned char SPI_GetManufacturerID(unsigned char chipno)
{
 	 unsigned char instatus,instatus2;
	 
	 //  Set PC4 or PD2 low - Flash CS
	 FlashChipSelect(chipno,LOW);

	 // Send Continuous read command
 	 SPI_MasterTransmit(0x9F);

	 instatus = SPI_MasterReceive();
	 printf("Device Id = %x ",instatus);
	 instatus = SPI_MasterReceive();
	 printf("%x ",instatus);
	 instatus = SPI_MasterReceive();
	 printf("%x ",instatus);
	 instatus = SPI_MasterReceive();
	 printf("%x\n\r",instatus);
	 instatus2 = SPI_MasterReceive();
	 
	 // Set PC4 or PD2 high - deselect Flash chip
	 FlashChipSelect(chipno,HIGH);
	 
	 return instatus;
}

//  Get serial flash chip status
unsigned char SPI_SerialFlashStatus(unsigned char chipno)
{
 	 unsigned char instatus,instatus2;
	 
	 //  Set PC4 or PD2 low - Flash CS
	 FlashChipSelect(chipno,LOW);

	 // Send Continuous read command
 	 SPI_MasterTransmit(SF_READ_STATUS);

	 instatus = SPI_MasterReceive();
	 instatus2 = SPI_MasterReceive();
	 
	 // Set PC4 or PD2 high - deselect Flash chip
	 FlashChipSelect(chipno,HIGH);
	 
	 return instatus;
}

//  SPI Flash continuous read command for AT45DB041B
//  The AT45DB041B has 2048 264byte pages
void SPI_StartReadFlashPage(long address)
{
 	 int i;
	 
	 // Send Continuous read command
 	 SPI_MasterTransmit(SF_READ_CONT);
	 
	 // Convert address to 24 bit address
	 ConvertFlashAddress(address);
	 for(i=0; i<3; i++)  SPI_MasterTransmit(FlashAddress[i]);
	 
	 //  32 dummy bits
	 for(i=0; i<4; i++)  SPI_MasterTransmit(0x00);
}

//  SPI Flash Buffer Read command for AT45DB041B
void SPI_ReadFlashBuffer(long address,int buffernumber)
{
 	 int i;
	 
	 // Send buffer read command
	 if(buffernumber==1) SPI_MasterTransmit(SF_READ_BUFFER1);
	 else SPI_MasterTransmit(SF_READ_BUFFER2);
	 
	 // Convert address to 24 bit address
	 ConvertFlashAddress(address);
	 for(i=0; i<3; i++)  SPI_MasterTransmit(FlashAddress[i]);
	 
	 //  8 dummy bits
	 SPI_MasterTransmit(0x00);
}

void DisplayExtendedAddress(unsigned int extendedaddress)
{
	puts(":02000004");
	PrintWordHex(extendedaddress);
	puts("FF");  // TBD - calculate correct crc
	putCRLF();
}

void DisplayIntelRecord(unsigned int recordaddress,unsigned char bytesonline)
{
 	 unsigned char i,byte,ascii[16],asciicnt,crcvalue,tempvalue;
	
	 crcvalue = 0;
	 putchar(':');
	 //  Output bytes on line
	 PrintHex(bytesonline);
	 crcvalue += bytesonline;
	 //  Output hex start address
	 PrintWordHex(recordaddress);
     tempvalue = (recordaddress&0xff00)>>8;
	 crcvalue += tempvalue;
	 tempvalue = recordaddress&0x00ff;
	 crcvalue += tempvalue;
	 //  Output record type 0
	 puts("00");
	 //  Output data
	 for(i=0; i<bytesonline; i++) {
	     byte = SPI_MasterReceive();
		 ascii[asciicnt++] = byte;
		 crcvalue += byte;
	     PrintHex(byte);
	     }
	 // Output CRC
	 PrintHex((~crcvalue)+0x01);
	 asciicnt = 0;
	 if(showascii==YES)  {
		 puts("  ");
		 for(i=0; i<bytesonline; i++) {
		     if(ascii[i]>0x1F&&ascii[i]<0x80)  putchar(ascii[i]);
			 else putchar('.');
			 }
		 }
	 putCRLF();  
}

//  Display P1 Blackfin Flash memory
//    Blackfin must be reset
void DisplaySerialFlash(long startaddress,long numbytes,char readmode)
{
 	 char firstrecord;
 	 unsigned char i,byte,bytesonlastline,ascii[16],asciicnt,crcvalue,tempvalue;
	 unsigned int extendedaddress,oldextendedaddress,recordaddress;
	 long address,endaddress;

	 //  Set PC4 low - Flash CS
	 FlashChipSelect(FLASHCHIP_1,LOW);  

	 if(readmode==READCONTINUOUS) {
	 	 //puts("Continuous Read\n\r");
	     SPI_StartReadFlashPage(startaddress);
		 } 
	 else if(readmode==READBUFFER) SPI_ReadFlashBuffer(startaddress,1);
	 
	 address = startaddress; 
	 endaddress = startaddress+numbytes;
	 oldextendedaddress = address/65536;
	 bytesonline = 16;	
	 firstrecord = YES;
	 
	 while(address<endaddress) {
	 	 extendedaddress = address/65536;
	     recordaddress = address - extendedaddress*65536; 
		 
		 // Output extended record if start address > 65536
		 if((firstrecord==YES)&&(extendedaddress>0))  {
			 //puts("-1-");
			 firstrecord = NO;
		     DisplayExtendedAddress(extendedaddress);
			 }	
			 
		 // Traversed 65536 boundary?			 
	     else if(extendedaddress>oldextendedaddress) {
			 //puts("-2-");
		     oldextendedaddress = extendedaddress;
			 DisplayExtendedAddress(extendedaddress);
		 	 }
		
		 // Generate current Hex record
		 else {
		     firstrecord = NO;
		     if(((long)recordaddress+bytesonline)>65536) {
				 //puts("-3-");
			     bytesonline = (65536 - recordaddress);
			     DisplayIntelRecord(recordaddress,bytesonline);
			     address += (long) bytesonline;
				 bytesonline = 16;
				 }
			 else if(((long)recordaddress+(long)bytesonline+(long)extendedaddress*65536)>endaddress) {
				 //puts("-3-3-");
			     bytesonline = endaddress - ((long)recordaddress+(long)extendedaddress*65536);
			     DisplayIntelRecord(recordaddress,bytesonline);
			     address += (long) bytesonline;
				 bytesonline = 16;
				 }
			 else {  
			     DisplayIntelRecord(recordaddress,bytesonline); 
			     address += (long) bytesonline; 
				 }
			 }
		}			 
		 
	 // Set PC4 high - deselect Flash chip
	 FlashChipSelect(FLASHCHIP_1,HIGH);  
	 puts(":00000001FF\n\r");
}

// Transfer buffer n to flash memory with built-in erase
//  This operation takes about 20msec max
void  ProgramBuffertoPageErase(long startaddress,unsigned char bufferid)
{
     unsigned char i;
	 
	 //puts("+p b p e+");
	 
	 // Serial Flash CS low
	 FlashChipSelect(FLASHCHIP_1,LOW);    
	 
	 //  Buffer write
	 if(bufferid==1) SPI_MasterTransmit(SF_BUF1_TO_FLASH_ERASE);
	 else SPI_MasterTransmit(SF_BUF2_TO_FLASH_ERASE);
	 
//	 puts(" write page at ");
	 // Convert address to 24 bit address
	 ConvertFlashAddress(startaddress);
	 
	 //FlashAddress[0] = 0x00;
	 for(i=0; i<3; i++)  SPI_MasterTransmit(FlashAddress[i]);
	 
	 // Serial Flash CS high - start erase
	 FlashChipSelect(FLASHCHIP_1,HIGH);   	 
}

// Maximum erase time is 20msec
void  EraseFlashPage(long startaddress)
{
     unsigned char i;
	 
	 //puts("+e f p+");

	 // Serial Flash CS low
	 FlashChipSelect(FLASHCHIP_1,LOW);    
	 
	 //  Buffer write
	 SPI_MasterTransmit(SF_PAGE_ERASE);
	 
	 // Convert address to 24 bit address
	 ConvertFlashAddress(startaddress);
	 for(i=0; i<3; i++)  SPI_MasterTransmit(FlashAddress[i]);
	 
	 // Serial Flash CS high - start erase
	 FlashChipSelect(FLASHCHIP_1,HIGH);   	 
}

//  Load data into flash chip RAM buffer
void ProgramFlashBuffer(long startaddress,unsigned char bufferid)
{
     unsigned char i;
	 
	 //puts("+p f b+");

	 // Serial Flash CS low
	 FlashChipSelect(FLASHCHIP_1,LOW);    
	 
	 //  Buffer write
	 if(bufferid==1)  SPI_MasterTransmit(SF_WRITE_BUFFER1);
	 else  SPI_MasterTransmit(SF_WRITE_BUFFER2);
	 
	 // Convert address to 24 bit address
	 ConvertFlashAddress(startaddress);
	 //FlashAddress[0] = 0x00;
	 for(i=0; i<3; i++)  SPI_MasterTransmit(FlashAddress[i]);
	 
	 for(i=0; i<bytesonline; i++)  SPI_MasterTransmit(intelbinary[i]);
	 
	 // Serial Flash CS high
	 FlashChipSelect(FLASHCHIP_1,HIGH);
}

void LoadFlashPage(long startaddress,unsigned char bufferid)
{
     unsigned char i;
	 
	 //puts("+l f p+");
	 // Serial Flash CS low
	 FlashChipSelect(FLASHCHIP_1,LOW);    
	 
	 //  Buffer write
	 if(bufferid==1)  SPI_MasterTransmit(SF_FLASH_TO_BUF1);
	 else  SPI_MasterTransmit(SF_FLASH_TO_BUF2);
	 
	 // Convert address to 24 bit address
	 ConvertFlashAddress(startaddress);
	 FlashAddress[0] = 0x00;
	 for(i=0; i<3; i++)  SPI_MasterTransmit(FlashAddress[i]);
	 
	 // Serial Flash CS high
	 FlashChipSelect(FLASHCHIP_1,HIGH);
    
}

void  WaitForFlashFree(unsigned char chipno)
{ 	  
	  while(!(SPI_SerialFlashStatus(chipno)&0x80)) ;
	  //Delaymsec(30);
}

int totalbytecnt;

//  Main Serial Flash programming state controller
//
//  Serial flash chips are programmed one page at a time.  A serial flash chip
//  has two SRAM buffers.  Due to AVR limitations programming data is first
//  loaded into the SRAM buffers then transferred to Flash.  SRAM buffers 
//  range in size from 264 to 1056 bytes so the state machine must keep track
//  of different buffer sizes.  IntelHex records range in size from 1 to 32
//  bytes (AVR limitation) and can arrive at any start address.
//
//  As IntelHex records are received on the serial channel they are first
//  loaded into Flash SRAM (AVR doesn't have enough SRAM available) and then
//  written to a Flash page if a termination record is received or the SRAM 
//  buffer becomes full.  To prevent Flash overwrite a Flash page is first 
//  loaded into SRAM and new data overwritten in SRAM.  
//
//  To reduce variable use bytesonline, inteladdress and intelsuperaddress
//  are global variables.
//
//  TBD - support for two flash chips
void ProgramSerialFlash(void)
{
 	 long globaladdress,shortaddress;
	 unsigned int newpage,newpage2,shortbytecnt,newbytecnt,i;
	 
	 globaladdress = intelextendedaddress*65536 + inteladdress;
	 //puts("=1="); PrintLongHex(globaladdress,6); puts("=1="); 
	 newpage = globaladdress/flashpagesize0;
	 
     // Always load flash page for first record
 	 if(flashprogrammingactive==NO) {
	     totalbytecnt = 0;
		 LoadFlashPage(globaladdress,1);
		 WaitForFlashFree(1);  // wait for transfer to complete - 300usec
	     flashprogrammingactive = YES;
		 flashpageloaded = YES;
		 oldpage = newpage;
		 }

 	 // See if flash page was loaded - preserves old data not in the new
	 // record from being overwritten
 	 else if(flashpageloaded==NO)  {
//	     puts(" load page ");
	     LoadFlashPage(globaladdress,1);
		 WaitForFlashFree(1);  // wait for programming to complete - 20msec
		 flashpageloaded = YES;
		 }
		 
	 // See if number of bytes in record causes a page boundary traverse
	 newbytecnt = bytesonline;
	 newpage2 = (globaladdress+bytesonline)/flashpagesize0;
	 if(newpage!=newpage2) {
		 shortaddress = newpage*flashpagesize0;	 
		 shortbytecnt = flashpagesize0 - (globaladdress - shortaddress);
		 newbytecnt = bytesonline - shortbytecnt;
		 bytesonline = shortbytecnt;
		 if(newbytecnt>0) flashpageboundarytraversed = YES;
//		 puts(" shortcnt="); printint(shortbytecnt,3,' ');
//		 puts(" newbytcnt="); printint(newbytecnt,3,' ');
		 }
	 
	 totalbytecnt += bytesonline;
	 ProgramFlashBuffer(globaladdress,1);  //puts(" l buf="); printint(totalbytecnt,5,' ');
	 flashdatainbuffer = YES;
	 
	 //  If Buffer full or boundary traversed transfer buffer to flash
	 if((newbytecnt==0)||(flashpageboundarytraversed==YES))  {
	     ProgramBuffertoPageErase(globaladdress,1);  //puts(" write page ");
		 WaitForFlashFree(1);  // wait for programming to complete - 20msec
		 flashpageloaded = NO;
		 flashdatainbuffer = NO;
//		 puts(" p page="); printint(newpage,3,' '); putchar(' ');
		 }
	 
	 //  If record caused a page boundary traversal load next page
	 if(flashpageboundarytraversed==YES) {
	     newpage++;       // goto next page
		 oldpage = newpage;
//		 puts(" t page="); printint(newpage,3,' '); putchar(' ');
		 globaladdress = newpage*flashpagesize0;  // start address at page bdry
		 lastaddress = globaladdress;
		 LoadFlashPage(globaladdress,1);   // Load the new page - this may be last record
		                     // So don't overwrite remaining content in page
		 WaitForFlashFree(1);
		 flashpageloaded = YES;  // Page loaded so don't do it again next recrd
		 for(i=0; i<shortbytecnt; i++)  // Shift record data to start
		      intelbinary[i] = intelbinary[shortbytecnt+i];
		 bytesonline = shortbytecnt;  // Only write new data in record
	 	 totalbytecnt += bytesonline;
		 ProgramFlashBuffer(globaladdress,1);  // write remaining record to RAM buffer
		 flashdatainbuffer = YES;
		 flashpageboundarytraversed = NO;
	     }
	 lastaddress = globaladdress;
}

// Return Serial Flash page size
//  Assume 321D and 642D are programmed as Power 2 devices
int FlashPageSize(unsigned char sizecode)
{
     //printf("Flash code = %x\n",sizecode);
 	 if((sizecode==AT45DB041ID)||(sizecode==AT45DB081ID))  return 264;
	 else if(sizecode==AT45DB161ID) return 528;
	 else if(sizecode==AT45DB321ID) return 512; // Must program as Power 2
	 else if(sizecode==AT45DB642ID) return 1024; // Must program as Power 2
	 else return 0;
}

void CheckPower2(unsigned char chipid,unsigned char status)
{
 	 if(status&0x01) return;  // If bit 0 set then Power 2 enabled 
	 if(((status&0x3C)==AT45DB321ID)||((status&0x3C)==AT45DB642ID)) {
	     printf("Program power 2 in chip %d\n\r",chipid);
	     ProgramPower2(chipid);
		 }
}

//  Search for serial flash chips U2, U5
//  U2 is always present while U5 is optional
//  This routine initializes SPI and resets Blackfin
void  CheckFlashTypes()
{
	Init_SPI_Master();
	ResetBlackfinBootMode(BOOTMODE0);
//	SPI_GetManufacturerID(FLASHCHIP_1);
	flashtype0 = SPI_SerialFlashStatus(FLASHCHIP_1);
    //printf("Flashtype0 = %x\n\r",flashtype0);
	if(flashtype0==0xff) flashpagesize0 = 0;
	else {
		CheckPower2(FLASHCHIP_1,flashtype0);  // Convert to Power 2
	    flashtype0 &= FLASHIDMASK;
	    flashpagesize0 = FlashPageSize(flashtype0);
		}
		
	flashtype1 = SPI_SerialFlashStatus(FLASHCHIP_2);
    //printf("Flashtype1 = %x\n\r",flashtype1);
	if(flashtype1==0xff) flashpagesize1 = 0;
	else  {
		CheckPower2(FLASHCHIP_2,flashtype1);  // Convert to Power 2
	    flashtype1 &= FLASHIDMASK;
	    flashpagesize1 = FlashPageSize(flashtype1);	
		}
	Release_SPI();
	ResetBlackfinBootMode(BOOTMODE3);
}

void EraseSerialFlash()
{
    int i;
	
	Release_SPI();
	ResetBlackfinBootMode(BOOTMODE0);
	Init_SPI_Master();	
	for(i=0; i<20; i++) {
	    //inteladdress = i*flashpagesize0;
		EraseFlashPage((long)(i*flashpagesize0));
		WaitForFlashFree(1);
		}
	Release_SPI();
	ResetBlackfinBootMode(BOOTMODE3);
}

