//*****************************************************************
//
//  LDRReader  - Parse a Blackfin LDR binary image and print record
//               headers
//
//  Version: 1.01    Date:  Dec 7, 2004
//  
//  A Blackfin LDR iamge consists of one or more blocks.  Each block
//  has a 10 header following by data.  The 10 byte header
//  contains a 32 bit start address, 32 bit byte count and 16 bit flag word.
//  The Start address defines the location in memory where the block is loaded.
//  The Count is the number of bytes in the data section of the block.
//
//  The program reads blocks until bit 15 of the Flag word is set indicating 
//  this is the last block.
//  
//  Flag Structure:
//    Bit 0 - ZEROFILL 0->non-zero fill block, 1->zero fill block
//    Bit 1 - RESVECT - 0->ADSP-BF532, 1->ADSP-BF533
//    Bit 3 - INIT - 0->Non-Init block, 1->Init block
//    Bit 4 - IGNORE - 0->Non-ignore block, 1->ignore block
//    Bit 8-5 - PFLAG 3:0 - PF number for SPI slave booting
//    Bit 15 - FINAL 0->non-last block, 1->Last block
//
//  Copyright 2004-2006, SOC Robotics, Inc.  All rights reserved.
//
//*****************************************************************
/*
 * Copyright (C) 2005-2006 by SOC Robotics, Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 * 4. Redistribution is only allowed in combination with SOC Machines, 
 *    Inc. hardware products.
 *
 * THIS SOFTWARE IS PROVIDED BY SOC ROBOTICS, INC. AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL OC MACHINES, 
 * INC. OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.soc-machines.com
 */

/*
 * $Log: LDRReader.c,v $
 * Revision 0.54  2005/06/04 18:36:42  steve
 * Preview release
 *
 */
 #include "P1_Main.h"
#include "ccblkfn.h"
#include "sysreg.h" 
#include "P1_Extern.h"

#define ZEROFILLBIT	0x0001
#define RESVECTBIT	0x0002
#define INITBIT		0x0008
#define IGNOREBIT	0x0010
#define PFLAGBITS	0x01E0
#define FINALBIT	0x8000



unsigned char ReadSDRAMByte(unsigned long address)
{
	return *(pSDRAMBot+address);
}

void BinaryFlag(unsigned short flagword)
{
	short i;

	PrintHexWord(flagword); 
	puts("  ");
	for(i=0; i<16; i++)  {
		if(i==8) putchar(' ');
		if(flagword&0x8000) putchar('1');
			else putchar('0');
		flagword <<= 1;
		}
}

unsigned long getlong()
{
	unsigned char inbyte;
	unsigned long word;

	inbyte = ReadSDRAMByte(addressptr++);
	word = inbyte + (ReadSDRAMByte(addressptr++)*256);
	word += (ReadSDRAMByte(addressptr++)*256*256);
	word += (ReadSDRAMByte(addressptr++)*256*256*256);
	return word;
}

unsigned int getword()
{
	unsigned char inbyte;
	unsigned short word;

	inbyte = ReadSDRAMByte(addressptr++);
	word = inbyte + (ReadSDRAMByte(addressptr++)*256);
	return word;
}

void LoadElfImage(unsigned long bootaddress)
{
	unsigned char crcvalue,intelbinary[16],inchar;
	char instrg[10],outstrg[10],infilename[80],outfilename[80];
	unsigned int  flagword;
	int  i,j,bytesonline,notfinished,bytecnt,blockcnt;
	unsigned long longaddress,longbytecount;

	longaddress = 0;
	blockcnt = 0;
	notfinished = 1;
	addressptr = bootaddress;

	//  Start output Hex data
	puts("\n\r");
	while(notfinished) {
		// Read Loader block header
		longaddress = getlong();
		longbytecount = getlong();
		flagword = getword();
		blockcnt++;

		// Output header data
		puts("Block "); printint(blockcnt,4,0); puts(" header information\n\r");
		puts("  Start address (hex,decimal): ");
		printlonghex(longaddress);
		puts("\n\r  Byte count (hex,decimal): ");
		printlonghex(longbytecount);
		puts("\n\r  Flag word: "); BinaryFlag(flagword);
		puts("\n\r    Zero fill[");  
		    if(flagword&ZEROFILLBIT) puts("1]");
			else puts("0]");
		puts("   Resvect[");
			if(flagword&RESVECTBIT)  puts("1]");
			else puts("0]");
		puts("  Init[");
			if(flagword&INITBIT)  puts("1]\n\r");
			else puts("0]\n\r");
		puts("    Ignore[");
			if(flagword&IGNOREBIT)  puts("1]");
			else puts("0]");
		puts("    PFlag PF"); 
			inchar = ((flagword&PFLAGBITS)>>5)&0x000F;
			PrintHexByte(inchar);
		puts("    Final bit ");
			if(flagword&FINALBIT)  puts("Set\n\r");
			else puts("Not Set\n\r");
		
		if(flagword&FINALBIT) notfinished = 0;  // Last block if Final Bit set

		
		// Read block data
		else {
			for(i=0; i<longbytecount; i++)  {
				*(pSDRAMBot+longaddress+i) = ReadSDRAMByte(addressptr++);
				inchar = ReadSDRAMByte(addressptr++);
				// Escape - hit any key
				if(getcharnw()) {
					notfinished = 0;
					break;
					}
				}

			}
		}
}

