/**************************************************************

   Program Title:  P1 AVR Boot/Monitor
   
   Program Name:  P1_AVR_Monitor.c	  Version: 0.87
   
   Author:  SS	  					  Date:  October 10, 2004
   
   Compiled with ImageCraft ICCAVR Standard V7.02
   
   The P1 AVR Boot Monitor is a set of functions that configure the 
   AVR running on the P1 processor for general operation and communication 
   with the P1 Blackfin processor.  Not all of these routines are required 
   for P1 operation but have been included to simplify custom application 
   development.   The AVR is a smart, programmable, IO engine for the Blackfin
   with TWI, analog and digital IO that can be monitored and controlled by 
   user applications running independently on the AVR.  
   
   The AVR Boot Monitor resets the Blackfin to boot mode 3 at power up allowing
   the Blackfin to boot from external serial flash.  If this code is modified
   the Blackfin control routines should not be modified or the Blackfin may
   be left in a reset state.  The Blackfin and AVR share the SPI lines so 
   any use of the AVR SPI functions should be coordinated with Blackfin use.
   The serial flash can be programmable by the AVR.  In this release the AVR 
   leaves the SPI port in default slave mode to support communication with 
   the Blackfin. 
   
   Key functions:
   	   Blackfin Boot Control
	   Blackfin Serial Flash programming/access
	   TWI 2-Wire Master Interrupt Controller
	   Real Time Clock setup with four configurable software timers
	   Full duplex interrupt drive USART with XON/XOFF flow control
	   1msec timer with four configurable software times
	   Analog controller setup
	   Digital port configuration
	   IntelHex32 download/upload for EEPROM, AVRFLASH and Serial Flash
	   RS-485 software UART configuration
	   Auto clock calibration
	   
   Standard printf and scanf functions are not used in this version 
   due to code size constraints.  The P1 is available with an ATmega32
   with twice the Flash and SRAM if printf and scanf functionality is 
   important or a custom application requires more than the available
   AVR Flash.
   
   Copyright 2004-2006, SOC Machines, Inc.  All rights reserved.
   
****************************************************************/
/*
 * Copyright (C) 2005 by SOC Machines, Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 * 4. Redistribution is only allowed in combination with SOC Machines, 
 *    Inc. hardware products.
 *
 * THIS SOFTWARE IS PROVIDED BY SOC MACHINES, INC. AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL OC MACHINES, 
 * INC. OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.soc-machines.com
 */

/*
 * $Log: P1_Monitor.c,v $
 * Revision 0.87  2005/07/17 16:40:20  SS
 *    Intel Hex 2 Record support, drop OK, osc calibration.
 * Revision 0.85  2005/07/17 16:40:20  SS
 *    2-wire interface, 8MHz clock, AVR Flash writer, 9,600baud
 *    real time clock.
 * Revision 0.52  2005/06/04 18:36:42  SS
 * Preview release
 * 
 */
#include <iom16v.h>
#include <macros.h>
#include <string.h>
#include <stdlib.h>
//#include <stdio.h>
#include "P1_AVR_Uart.h"
#include "P1_AVR_Defines.h"
#include "P1_AVR_Extern.h"

//  Variable declarations
unsigned int bytesonline,inteladdress,intelextendedaddress;
long  lastaddress;
unsigned char programmode,crc,intelbinary[48];  // length can vary
unsigned char hexbyte[3],hexaddress[5],intelrecordcommand;
unsigned char seconds,minutes,hours,days,rtcclockcnt;
unsigned char flashtype0,flashtype1,flashpageloaded,flashpageboundarytraversed;
unsigned char flashprogrammingactive,flashdatainbuffer;
unsigned char xofftransmitted,showascii;
int	flashpagesize0,flashpagesize1,startaddress,currrentaddress;
int oldpage,oldbyteoffset,totalrxcnt;
int rtccountdown1,rtccountdown2,rtccountdown3,rtccountdown4;
int mseccountdown1,mseccountdown2,mseccountdown3,flashcountdown;

extern int _textmode;

/* Static Variables */
unsigned char UART_RxBuf[UART_RX_BUFFER_SIZE];
unsigned char UART_RxHead;
unsigned char UART_RxTail;
unsigned char UART_TxBuf[UART_TX_BUFFER_SIZE];
unsigned char UART_TxHead;
unsigned char UART_TxTail;
char  xoffsent, xonsent, enablexonoff, echoinchars, sendxonxoff;
char  rxcnt,checkflag,rxcntmax;
unsigned char  ip[5],ipmask[5],ipgate[5],instrg[8];

unsigned char FlashAddress[3];
const char signon1[] = "\n\r\n\rP1 Pirana AVR Monitor V0.87\n\r";
const char signon2[] = "Copyright 2004-2006,  SOC Machines, Inc.\n\r\n\r";
const char signon3[] = "   D[E|A|F]  - Display AVR EEPROM, AVR Flash or Flash memory\n\r";
const char signon4[] = "   P[E|A|F]  - Program AVR EERPOM, AVR Flash or Flash memory\n\r";
const char signon5[] = "   R   - Reset Blackfin - Blackfin released\n\r";
const char signon6[] = "   H   - Halt/Release Blackfin (Red LED on - Blackfin Halted)\n\r";
const char signon65[] = "   B   - Set Blackfin Bootmode [0,2,3] 3-default\n\r";
const char signon66[] = "\n\r Enter Blackfin Bootmode [0,2,3]: ";
const char signon67[] = "\n\r Blackfin Bootmode set to ";
const char signon7[] = "   V   - Version\n\r";
const char signon8[] = "   C   - Test AVR\n\r";
const char signon83[] = "   T   - Get/Set Time\n\r";
const char signon85[] = "   A   - Read all analog channels\n\r";
const char signon9[] = "   S   - Signon message\n\r";
const char signon10[] = "  Note:  Blackfin held in reset until released by R command\n\r\n\r-";
const char signon11[] = "   Serial flash 0 page size = ";
const char signon12[] = "   Serial flash 1 page size = ";
const char crlfbar[] = "\n\r-";

const char displaymessage1[] = "Enter start address and length: ";
const char displaymessage2[] = "\n\rDisplay AVR Flash from 0x0054 to 254\n\r";
const char displaymessage3[] = "\n\rDisplay SPI Flash from 0 to 100\n\r";
const char displaymessage33[] = "\n\rNo mode selected\n\r";

const char progmessage[] = "P1 Pirana AVR Monitor Version: ";

const char displaymessage4[] = "\n\rLoad EEPROM from address ";
const char displaymessage5[] = " length ";
const char displaymessage6[] = "\n\rLoad SPI Flash from address ";

const char loadmessage1[] = "\n\rLoad EEPROM\n\r";
const char loadmessage2[] = "\n\rLoad SPI Flash\n\r";
const char loadmessage3[] = "\n\rLoad AVR Flash\n\r";

// Function declarations
unsigned char EEPROMwrite(int,unsigned char);
unsigned char EEPROMread(int);
int printf(CONST char *, ...);
int sprintf(char *, CONST char *, ...);
int scanf(CONST char *, ...);
int sscanf(char *, CONST char *, ...);

// Motor control Commands
#define    MOTORBOTHFORWARD  	'1'    //    FB - forward both motors
#define    MOTORRIGHTFORWARD  	'2'    //    FR - forward right motor
#define    MOTORLEFTFORWARD  	'3'    //    FL - forward left motor
#define    MOTORBOTHREVERSE  	'4'    //    RB - reverse both motors
#define    MOTORREVERSERIGHT  	'5'    //    RR - reverse right motor
#define    MOTORREVERSELEFT  	'6'    //    RL - reverse left motor
#define    ROTATECW				'c'	   //    Rotate clockwise
#define    ROTATECCW			'w'    //    Rotate counter clockwise
//    Axxx - Acceleration rate xxx - 000 to 999
#define    MOTORACCELERATION  	'7'    
#define    MOTORHALT  	        'h'    //    H - Halt 
#define    LIGHTCONTROL  	    '9'    //    LxH - turn light x[O]N/O[F]F
//    Sxxx - set speed to xxx - 000 to 999 (relative as ratio)
#define    MOTORRIGHTSPEED      'a'
//    Sxxx - set speed to xxx - 000 to 999 (relative as ratio)    
#define    MOTORLEFTSPEED       'b'    
#define    RESETCONTROLLER  	'g'    //    E - reset to defaults

void Sumo1_Pattern(void)
{
    unsigned char pattern[20] = {'1','2','3','5','6','4','c','w',
			 	  			  	 '1','2','3','5','6','4','c','w',
								 '1','4','c','w'};
	char inchar;
	int i;
	
	//puts("\n\rDo repetitive pattern - press any key to stop");
	while(!(inchar=getcharnw())) {
	    for(i=0; i<20; i++) {
	        TWI_Transmit(0x15,pattern[i],QUIET);
		    LongDelay(8);
		    }
		}
	TWI_Transmit(0x15,'h',QUIET);
}

void Sumo1_Controller(void)
{	 
	unsigned char inchar;
	
    //puts("Sumo 1 Command Controller");
	//puts("\n\rEnter command[1,2,3,4,5,6,h,c,w][e to exit]: ");
 	while((inchar=getchar())!='e')  {
	    putchar(inchar);
		if(inchar=='p') Sumo1_Pattern();
		else {
		    if(TWI_Transmit(0x15,inchar,QUIET)!=0x01) {
		        TWI_Off();
		        Init_TWI_Master();
		        }
			}
		//puts("\n\rEnter command[1,2,3,4,5,6,h,c,w][e to exit]: ");
		}
}

//*************************************************************
//  Main work starts here
//*************************************************************
void main()
{
 	char toggle,slaverx;
	unsigned char inchar,inbyte,twistatus1,twistatus2,twistatus3,twistatus4;
	unsigned char inbytes[5];
	int i,icnt,j,loopsetting,loopmode,blackfinresetmode,countervalue,numbr;
	
	_textmode = 0;  // do not expand /n to CRLF
	Init_AVR();
	rxcntmax = totalrxcnt = 0;
	
	loopsetting = 1000;   // 1 second delay
	loopmode = 0;  toggle = 0;
	mseccountdown1 = loopsetting;

	// Get Serial Flash status
	CheckFlashTypes();
	// Reset and release Blackfin
	ResetBlackfinBootMode(BOOTMODE3);
	ReleaseBlackfin(QUIET);	
	
	// Show something is happening
	BlinkRed(2);
	BlinkGreen(2);
	signon();
	
	// Debug - clean out the PC serial driver
	// Remove when serial driver reliable
//	for(i=0; i<10; i++)  putchar(XON);
	
	blackfinresetmode = 0x00;
//	WriteEEPROM(0,"SOC Machines V0.87");

	// Wait for commands on serial channel
	while (1) {
		  // Work starts here
		  if((inchar = getcharnw())) {
			   	switch (inchar) {
					 case 's': case 'S':
						 signon();
						 //puts("rxcnt="); printint(rxcnt,3,' ');
						 //if(xofftransmitted==YES)  puts(" xoff sent");
						 //else puts("xoff not sent"); 
						 //putCRLF(); putchar('-');
						 break;
					    case 'd': case 'D':
						    putchar(inchar);
						    DisplayMemory();
							putchar('-');
							break;
					    case 'p': case 'P':
							putchar(inchar);
						    LoadMemory();
							putchar('-');
							break;
					    case 'a': case 'A':
						    DumpAnalogChs();
							putchar('-');
							break;
					    case 'v': case 'V':
						    ProgramVersion();
							putchar('-');
							break;
					    case 'r': case 'R':
							Release_SPI();
						    ResetBlackfin(VERBOSE);
							putchar('-');
							break;
					    case 'b': case 'B':
							putchar(inchar);
							SetBlackfinBootMode();
							putchar('-');
							break;
					    case 'h': case 'H':
							if(blackfinresetmode==0x01) {
								 ReleaseBlackfin(VERBOSE);
								 blackfinresetmode = 0x00;
								 }
							else {
								 HaltBlackfin(VERBOSE);
								 blackfinresetmode = 0x01;
								 }
							putchar('-');
							break;
					    case ':':
						    ProgramMemory();
							//putchar('-');
							break;
						case 'i': case 'I':
						    InternetSettings();
							putchar('-');
							break;
						case 'e': case 'E':
						    EraseSerialFlash();
							puts("Erase\n\r");
							putchar('-');
							break;
						case 'l': case 'L':
						    while(!(inchar=getcharnw())) putchar(0x55);
							putCRLF();
							putchar('-');
							break;
						case 'c': case 'C':
						    while(!(inchar=getcharnw()))  TestAVR();
							signon();
							break;
						case 't': case 'T':
						    getsettime();
							putCRLF();
							putchar('-');
							break;
						case 'x': case 'X':
							putchar(inchar);
							TWI_Write_Slave();
							putchar('-');
							break;
						case 'q': case 'Q':
							putchar(inchar);  //puts("\n\rTWI Master Read Data ");
						    // while(!(inchar=getcharnw()))  TestTWI();
							Init_TWI_Master();
							inbyte = TWI_Master_Receive(0x15);
						    //while(!(inchar=getcharnw()))  {
							    // Send start command and wait for completion
								/*TWCR = (1<<TWINT)|(1<<TWEN)|(1<<TWSTA);
								while(!(TWCR & (1<<TWINT)))  ;	
								twistatus1 = TWSR;
								// Send slave address and read command 
								TWDR = (0x15<<1) | SLA_R;
								TWCR = (1<<TWINT)|(1<<TWEN);
								while(!(TWCR & (1<<TWINT)))  ;  
								twistatus2 = TWSR; 
								// Read data byte and wait for transmission to complete
	  							TWCR = (1<<TWINT)|(1<<TWEA)|(1<<TWEN);
								countervalue = 0;
	  							while(!(TWCR & (1<<TWINT))) {
									if(countervalue++>40) break;
									} 	
								inbyte = TWDR;    
								twistatus3 = TWSR;
								// Read next data byte and wait for transmission to complete
	  							TWCR = (1<<TWINT)|(1<<TWEN);
								countervalue = 0;
	  							while(!(TWCR & (1<<TWINT))) {
									if(countervalue++>40) break;
									} 	
								//inbyte = TWDR;    
								twistatus4 = TWSR;
								
								if(countervalue>40) {
								   TWI_Off();
								   Init_TWI_Master();
								   }
								else TWCR = (1<<TWINT)|(1<<TWEN)|(1<<TWSTO);
								shortdelay(5);  
								//}
							    TWI_Off();
							    printf("\n\rTWI status bytes = %x %x %x %x"
								  twistatus1,twistatus2,twistatus3,twistatus4);
								*/
								TWI_Off();
							    //printf("\n\rTWI Master Read byte = %x  %c\n\r-",inbyte,inbyte); 
							    break;
							case 'y': case 'Y':
						        // This version assumes single byte reception
							    putchar(inchar);  puts(" TWI Slave Active ");
						        // while(!(inchar=getcharnw()))  TestTWI();
							    slaverx = 1;  icnt = 0;
							    Init_TWI_Slave(0x15,TWI_GENERAL_CALL_ON);
						        while(!(inchar=getcharnw()))  {
							        // If slave address match process data
								    // otherwise loop and exit if keyboard key pressed
								    if((TWCR & (1<<TWINT))) {
								        for(i=0; i<5; i++)  inbytes[i] = 0;
									    i = 0;	slaverx = 2;
								        // Clear flag and wait for data reception
									
									    // If data receive command - read until
									    //   stop condition received
									    if((TWSR&0xF8)==SR_WRITE_COMMAND) {
								            TWCR = (1<<TWINT)|(1<<TWEA)|(1<<TWEN);
								            while(!(TWCR & (1<<TWINT)))  ;
									
									        // Read data (up to 5 bytes) 
									        //   until stop condition detected
									        //   TWEA send ACK
								            while((TWSR&0xF8)!=TWI_STOP)  {
										        if(i>=5) break;
									            inbytes[i++] = TWDR; 
										        icnt = i;     			 
	  							                TWCR = (1<<TWINT)|(1<<TWEA)|(1<<TWEN);
	  							                while(!(TWCR & (1<<TWINT)))  ; 	  
								                TWCR = (1<<TWINT)|(1<<TWEA)|(1<<TWEN);
								                } 
										    }
									    // If data send command - send data with
									    //   TWEA set until last byte 
									    //   this example sends only one byte 
									    else if((TWSR&0xF8)==SR_READ_COMMAND) {
									    //else  {
									        putchar('v');
									        TWDR = 0x78;  // data to send
										    TWCR = (1<<TWINT)|(1<<TWEN);
										    //TWCR = (1<<TWINT)|(1<<TWEN);
										    while(!(TWCR & (1<<TWINT)))  ;
										    TWCR = (1<<TWINT)|(1<<TWEA)|(1<<TWEN); 
										    }
	 								    }
								    }
							TWI_Off();
							if(slaverx==2) {
								puts("\n\r"); 
							    printint(icnt,2,' '); 
								//puts(" Slave data bytes rx'd = "); 
							    for(i=0; i<icnt; i++) {
								    bytetohex(inbytes[i]);
								    puts(hexbyte); putchar(' ');
									}
								for(i=0; i<icnt; i++) {
								    putchar(inbytes[i]);
									}
								}
							else  ;//puts("\n\rNo Slave data rx'd");
							//shortdelay(6);
							//TWCR = (1<<TWINT)|(1<<TWEN);
							puts("\n\r-");
							break;
						case LF:
							putCRLF();
							putchar('-');	
							break;
						case CR:
							putCRLF();
							putchar('-');	
							break;
						default:
							putchar(inchar);
							//Delaymsec(80);
							break;
						}
			   		
			   		} 

			   // Blink green led slowly to show AVR is alive
			   if(mseccountdown1==0) {
			   		mseccountdown1 = loopsetting;
					if(loopmode)  {
					    GREEN_LED(ON);
						loopmode = 0;
						}
					else {
						GREEN_LED(OFF);
						loopmode = 1;
						}
			   	 	}
					
				// Clean up flash state machine if no new Intel
				// Hex records for 3 seconds
				if((flashcountdown==0)&&(flashprogrammingactive==YES))
				        CleanUpFlashState();
		}
}
